"use strict";
/**
 * A Lexical analyser
 * @module lexer
 */
Object.defineProperty(exports, "__esModule", { value: true });
/**
 * A Lexer instance, parsing a given file.  Usually you should use a LexicalGrammar to
 * create one of these.
 *
 * @class
 * @param {string} source the source code to parse
 * @param rules the rules of this lexer.
 */
class Lexer {
    constructor(source, rules) {
        this.source = source;
        this.rules = rules;
        this.position = 0;
    }
    peek() {
        return this.peeked = this.scan();
    }
    match(type, raw) {
        let p = this.peek();
        if (p && p.type == type && (!raw || p.raw == raw)) {
            this.peeked = null;
            return true;
        }
        return false;
    }
    scan() {
        if (this.peeked) {
            let res = this.peeked;
            this.peeked = null;
            return res;
        }
        var token = null;
        var length = 0;
        this.rules.forEach(rule => {
            rule.r.lastIndex = this.position;
            var x = rule.r.exec(this.source);
            if (x && x[0].length > length && this.position + x[0].length == rule.r.lastIndex) {
                token = rule.fn(this, x);
                token.offset = this.position;
                token.raw = x[0];
                length = x[0].length;
            }
        });
        this.position += length;
        if (token == null) {
            if (this.position == this.source.length)
                return null;
            throw new Error("Unexpected character at " + this.position + ": " + JSON.stringify(this.source));
        }
        return token;
    }
}
exports.Lexer = Lexer;
/**
 * A lexical grammar- factory for lexer instances.
 * @class
 */
class LexicalGrammar {
    constructor() {
        this.rules = [];
    }
    /**
     * Defines a terminal with the given pattern and constructor.
     * @param {string} pattern the pattern this nonterminal must match.
     * @param {function(Array<string>): Object} fn returns a lexical token representing
     *        this terminal.  An additional "offset" property containing the token source position
     *        will also be added, as well as a "raw" property, containing the raw string match.
     */
    terminal(pattern, fn) {
        this.rules.push({ r: new RegExp(pattern, "g"), fn: fn });
    }
    lex(source) {
        return new Lexer(source, this.rules);
    }
}
exports.LexicalGrammar = LexicalGrammar;
//# sourceMappingURL=lexer.js.map