"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const vscode = require("vscode");
const _ = require("lodash");
const state = require("../../state");
const evaluate_1 = require("./evaluate");
const util = require("../../utilities");
let diagnosticCollection = vscode.languages.createDiagnosticCollection('calva');
function reportTests(results, errorStr, log = true) {
    let chan = state.deref().get('outputChannel'), diagnostics = {}, total_summary = { test: 0, error: 0, ns: 0, var: 0, fail: 0 };
    diagnosticCollection.clear();
    if (results.err || results.ex) {
        util.logError({
            type: util.ERROR_TYPE.ERROR,
            reason: "Error " + errorStr + ":" + results.err
        });
    }
    else {
        for (let result of results) {
            for (const ns in result.results) {
                let resultSet = result.results[ns];
                for (const test in resultSet) {
                    for (const a of resultSet[test]) {
                        if (a.type == "error" && log)
                            chan.appendLine(`ERROR in: ${ns}: ${a.file}, line ${a.line}: ${test}: ${(a.context || "")}:\n  error: ${a.error} + "\n  expected: ${a.expected}`);
                        if (a.type == "fail") {
                            let msg = `failure in test: ${test} context: ${a.context}, expected ${a.expected}, got: ${a.actual}`, err = new vscode.Diagnostic(new vscode.Range(a.line - 1, 0, a.line - 1, 1000), msg, vscode.DiagnosticSeverity.Error);
                            if (!diagnostics[a.file])
                                diagnostics[a.file] = [];
                            diagnostics[a.file].push(err);
                            if (log)
                                chan.appendLine(`FAIL in: ${a.file}: ${a.line}: ${test}: ${(a.context || "")}:\n  expected: ${a.expected}\n  actual: ${a.actual}`);
                        }
                    }
                }
            }
            if (result.summary !== null) {
                _.each(result.summary, (v, k) => {
                    total_summary[k] = result.summary[k] + (total_summary[k] !== undefined ? total_summary[k] : 0);
                });
            }
        }
        if (total_summary !== null) {
            let hasProblems = total_summary.error + total_summary.fail > 0;
            if (log) {
                chan.appendLine("\n" + (total_summary.test > 0 ?
                    total_summary.test + " tests finished, " +
                        (!hasProblems ? "all passing 👍" :
                            "problems found. 😭" +
                                " errors: " + total_summary.error + ", failures: " + total_summary.fail) : "No tests found. 😱") +
                    ", ns: " + total_summary.ns + ", vars: " + total_summary.var);
            }
            if (total_summary.test > 0) {
                if (hasProblems) {
                    _.each(diagnostics, (errors, fileName) => {
                        if (fileName.startsWith('/'))
                            diagnosticCollection.set(vscode.Uri.file(fileName), errors);
                        else {
                            // Sometimes we don't get the full path for some reason. (This is a very inexact
                            // way of dealing with that. Maybe check for the right `ns`in the file?)
                            vscode.workspace.findFiles('**/' + fileName, undefined).then((uri) => {
                                diagnosticCollection.set(uri[0], errors);
                            });
                        }
                    });
                }
            }
        }
    }
}
// FIXME: use cljs session where necessary
function runAllTests(document = {}) {
    return __awaiter(this, void 0, void 0, function* () {
        let client = util.getSession(util.getFileType(document));
        state.deref().get('outputChannel').appendLine("Running all project tests…");
        reportTests([yield client.testAll()], "Running all tests");
    });
}
function runAllTestsCommand() {
    state.deref().get('outputChannel').show(true);
    runAllTests();
}
function considerTestNS(ns, client, nss) {
    return __awaiter(this, void 0, void 0, function* () {
        if (!ns.endsWith('-test')) {
            let testNS = ns + '-test', testFilePath = yield client.nsPath(testNS).path;
            if (`${testFilePath}` != "") {
                let loadForms = `(load-file "${testFilePath}")`;
                yield client.eval(loadForms);
            }
            nss.push(testNS);
            return nss;
        }
        return nss;
    });
}
function runNamespaceTests(document = {}) {
    return __awaiter(this, void 0, void 0, function* () {
        let client = util.getSession(util.getFileType(document)), doc = util.getDocument(document), ns = util.getNamespace(doc.getText()), nss = [ns];
        evaluate_1.default.evaluateFile({}, () => __awaiter(this, void 0, void 0, function* () {
            state.deref().get('outputChannel').appendLine("Running namespace tests…");
            nss = yield considerTestNS(ns, client, nss);
            let resultPromises = [client.test(nss[0])];
            if (nss.length > 1)
                resultPromises.push(client.test(nss[1]));
            let results = yield Promise.all(resultPromises);
            reportTests(results, "Running tests");
        }));
    });
}
function runNamespaceTestsCommand() {
    state.deref().get('outputChannel').show(true);
    runNamespaceTests();
}
function rerunTests(document = {}) {
    let client = util.getSession(util.getFileType(document));
    evaluate_1.default.evaluateFile({}, () => __awaiter(this, void 0, void 0, function* () {
        state.deref().get('outputChannel').appendLine("Running previously failed tests…");
        reportTests([yield client.retest()], "Retesting");
    }));
}
function rerunTestsCommand() {
    state.deref().get('outputChannel').show(true);
    rerunTests();
}
exports.default = {
    runNamespaceTests,
    runNamespaceTestsCommand,
    runAllTests,
    runAllTestsCommand,
    rerunTestsCommand
};
//# sourceMappingURL=testRunner.js.map