"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const vscode = require("vscode");
const specialWords = ['-', '+', '/', '*']; //TODO: Add more here
exports.specialWords = specialWords;
const _ = require("lodash");
const state = require("./state");
const fs = require("fs");
const syntaxQuoteSymbol = "`";
function getProjectDir() {
    let path = vscode.workspace.rootPath + "/" + state.config().projectRootDirectory;
    if (fs.existsSync(path)) {
        return path;
    }
    else {
        return vscode.workspace.rootPath;
    }
}
exports.getProjectDir = getProjectDir;
function getCljsReplStartCode() {
    return vscode.workspace.getConfiguration('calva').startCLJSREPLCommand;
}
exports.getCljsReplStartCode = getCljsReplStartCode;
function getShadowCljsReplStartCode(build) {
    return '(shadow.cljs.devtools.api/nrepl-select ' + build + ')';
}
exports.getShadowCljsReplStartCode = getShadowCljsReplStartCode;
function getNamespace(text) {
    let match = text.match(/^[\s\t]*(?:;.*\s)*[\s\t]*\((?:[\s\t\n]*(?:in-){0,1}ns)[\s\t\n]+'?([\w.\-\/]+)[\s\S]*\)[\s\S]*/);
    return match ? match[1] : 'user';
}
exports.getNamespace = getNamespace;
function getStartExpression(text) {
    let match = text.match(/^\(([^\)]+)[\)]+/g);
    return match ? match[0] : "(ns user)";
}
exports.getStartExpression = getStartExpression;
function getActualWord(document, position, selected, word) {
    if (selected === undefined) {
        let selectedChar = document.lineAt(position.line).text.slice(position.character, position.character + 1), isFn = document.lineAt(position.line).text.slice(position.character - 1, position.character) === "(";
        if (selectedChar !== undefined &&
            specialWords.indexOf(selectedChar) !== -1 &&
            isFn) {
            return selectedChar;
        }
        else {
            return "";
        }
    }
    else {
        return (word && word.startsWith(syntaxQuoteSymbol)) ? word.substr(1) : word;
    }
}
function getWordAtPosition(document, position) {
    let selected = document.getWordRangeAtPosition(position), selectedText = selected !== undefined ? document.getText(new vscode.Range(selected.start, selected.end)) : "", text = getActualWord(document, position, selected, selectedText);
    return text;
}
exports.getWordAtPosition = getWordAtPosition;
function getDocument(document) {
    if (document && document.hasOwnProperty('fileName')) {
        return document;
    }
    else if (vscode.window.activeTextEditor) {
        return vscode.window.activeTextEditor.document;
    }
    else if (vscode.window.visibleTextEditors.length > 0) {
        return vscode.window.visibleTextEditors[0].document;
    }
    else {
        return null;
    }
}
exports.getDocument = getDocument;
function getFileType(document) {
    let doc = getDocument(document);
    if (doc) {
        return doc.fileName.substr((doc.fileName.lastIndexOf('.') + 1), doc.fileName.length);
    }
    else {
        return 'clj';
    }
}
exports.getFileType = getFileType;
function getFileName(document) {
    let fileNameIndex = (document.fileName.lastIndexOf('\\') + 1);
    return document.fileName.substr(fileNameIndex, document.fileName.length);
}
exports.getFileName = getFileName;
function getDocumentNamespace(document = {}) {
    let doc = getDocument(document);
    return getNamespace(doc.getText());
}
exports.getDocumentNamespace = getDocumentNamespace;
function getSession(fileType = undefined) {
    let doc = getDocument({}), current = state.deref();
    if (fileType === undefined) {
        fileType = getFileType(doc);
    }
    if (fileType.match(/^clj[sc]?/)) {
        return current.get(fileType);
    }
    else {
        return current.get('clj');
    }
}
exports.getSession = getSession;
// ERROR HELPERS
const ERROR_TYPE = {
    WARNING: "warning",
    ERROR: "error"
};
exports.ERROR_TYPE = ERROR_TYPE;
function logSuccess(results) {
    let chan = state.deref().get('outputChannel');
    chan.appendLine("Evaluation completed successfully");
    _.each(results, (r) => {
        let value = r.hasOwnProperty("value") ? r.value : null;
        let out = r.hasOwnProperty("out") ? r.out : null;
        if (value !== null) {
            chan.appendLine("=>\n" + value);
        }
        if (out !== null) {
            chan.appendLine("out:\n" + out);
        }
    });
}
exports.logSuccess = logSuccess;
function logError(error) {
    let chan = state.deref().get('outputChannel');
    chan.appendLine(error.reason);
    if (error.line !== undefined && error.line !== null &&
        error.column !== undefined && error.column !== null) {
        chan.appendLine("at line: " + error.line + " and column: " + error.column);
    }
}
exports.logError = logError;
function markError(error) {
    if (error.line === null) {
        error.line = 0;
    }
    if (error.column === null) {
        error.column = 0;
    }
    let diagnostic = state.deref().get('diagnosticCollection'), editor = vscode.window.activeTextEditor;
    //editor.selection = new vscode.Selection(position, position);
    let line = error.line - 1, column = error.column, lineLength = editor.document.lineAt(line).text.length, lineText = editor.document.lineAt(line).text.substring(column, lineLength), firstWordStart = column + lineText.indexOf(" "), existing = diagnostic.get(editor.document.uri), err = new vscode.Diagnostic(new vscode.Range(line, column, line, firstWordStart), error.reason, vscode.DiagnosticSeverity.Error);
    let errors = (existing !== undefined && existing.length > 0) ? [...existing, err] :
        [err];
    diagnostic.set(editor.document.uri, errors);
}
exports.markError = markError;
function logWarning(warning) {
    let chan = state.deref().get('outputChannel');
    chan.appendLine(warning.reason);
    if (warning.line !== null) {
        if (warning.column !== null) {
            chan.appendLine("at line: " + warning.line + " and column: " + warning.column);
        }
        else {
            chan.appendLine("at line: " + warning.line);
        }
    }
}
exports.logWarning = logWarning;
function markWarning(warning) {
    if (warning.line === null) {
        warning.line = 0;
    }
    if (warning.column === null) {
        warning.column = 0;
    }
    let diagnostic = state.deref().get('diagnosticCollection'), editor = vscode.window.activeTextEditor;
    //editor.selection = new vscode.Selection(position, position);
    let line = Math.max(0, (warning.line - 1)), column = warning.column, lineLength = editor.document.lineAt(line).text.length, existing = diagnostic.get(editor.document.uri), warn = new vscode.Diagnostic(new vscode.Range(line, column, line, lineLength), warning.reason, vscode.DiagnosticSeverity.Warning);
    let warnings = (existing !== undefined && existing.length > 0) ? [...existing, warn] :
        [warn];
    diagnostic.set(editor.document.uri, warnings);
}
exports.markWarning = markWarning;
function updateREPLSessionType() {
    let current = state.deref(), doc = getDocument({}), fileType = getFileType(doc);
    if (current.get('connected')) {
        if (fileType == 'cljs' && getSession('cljs') !== null) {
            state.cursor.set('current-session-type', 'cljs');
        }
        else if (fileType == 'clj' && getSession('clj') !== null) {
            state.cursor.set('current-session-type', 'clj');
        }
        else if (fileType == 'cljc' && getSession('cljc') !== null) {
            state.cursor.set('current-session-type', getSession('cljc') == getSession('clj') ? 'clj' : 'cljs');
        }
        else {
            state.cursor.set('current-session-type', 'clj');
        }
    }
    else {
        state.cursor.set('current-session-type', null);
    }
}
exports.updateREPLSessionType = updateREPLSessionType;
function getREPLSessionType() {
    let current = state.deref();
    return current.get('current-session-type');
}
exports.getREPLSessionType = getREPLSessionType;
//# sourceMappingURL=utilities.js.map