"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const net = require("net");
const bencode_1 = require("./bencode");
/** An nRREPL client */
class NReplClient {
    constructor(socket) {
        this._nextId = 0;
        this.encoder = new bencode_1.BEncoderStream();
        this.decoder = new bencode_1.BDecoderStream();
        /** Tracks all sessions */
        this.sessions = {};
        this.ns = "user";
        this._closeHandlers = [];
        this.socket = socket;
        this.socket.on("error", e => {
            console.error(e);
        });
        this.socket.on("close", e => {
            console.log("Socket closed");
            this._closeHandlers.forEach(x => x(this));
            for (let x in this.sessions) {
                this.sessions[x]._onCloseHandlers.forEach(s => s(this.sessions[x]));
            }
        });
        this.encoder.pipe(this.socket);
        this.socket.pipe(this.decoder);
    }
    /** Returns a new id unique to this client */
    get nextId() {
        return ++this._nextId + "";
    }
    addOnCloseHandler(fn) {
        if (this._closeHandlers.indexOf(fn) == -1)
            this._closeHandlers.push(fn);
    }
    removeOnCloseHandler(fn) {
        let idx = this._closeHandlers.indexOf(fn);
        if (idx != -1)
            this._closeHandlers.splice(idx, 1);
    }
    /**
     * Returns a new session.
     */
    createSession() {
        return this.session.clone();
    }
    /**
     * Send a Javascript object over the wire as bencode.
     * @param data
     */
    write(data) {
        this.encoder.write(data);
    }
    close() {
        this.socket.destroy();
    }
    /**
     * Create a new NRepl client
     *
     * TODO - should just be a toplevel nrepl.createClient()
     */
    static create(opts) {
        return new Promise((resolve, reject) => {
            let socket = net.createConnection(opts, () => {
                let nsId = client.nextId;
                let cloneId = client.nextId;
                let describeId = client.nextId;
                client.decoder.on("data", data => {
                    //console.log("-> ", data);
                    if (!client.describe && data["id"] == describeId) {
                        client.describe = data;
                    }
                    else if (data["id"] == nsId) {
                        if (data["ns"])
                            client.ns = data["ns"];
                        if (data["status"] && data["status"].indexOf("done") != -1)
                            client.encoder.write({ "op": "clone", "id": cloneId });
                    }
                    else if (data["id"] == cloneId) {
                        client.session = new NReplSession(data["new-session"], client);
                        client.encoder.write({ "op": "describe", id: describeId, verbose: true, session: data["new-session"] });
                        resolve(client);
                    }
                    else if (data["session"]) {
                        let session = client.sessions[data["session"]];
                        if (session)
                            session._response(data);
                    }
                });
                client.encoder.write({ "op": "eval", code: "*ns*", "id": nsId });
            });
            let client = new NReplClient(socket);
        });
    }
}
exports.NReplClient = NReplClient;
class NReplSession {
    constructor(sessionId, client) {
        this.sessionId = sessionId;
        this.client = client;
        this._onCloseHandlers = [];
        this.messageHandlers = {};
        client.sessions[sessionId] = this;
    }
    addOnCloseHandler(fn) {
        if (this._onCloseHandlers.indexOf(fn) == -1)
            this._onCloseHandlers.push(fn);
    }
    removeOnCloseHandler(fn) {
        let idx = this._onCloseHandlers.indexOf(fn);
        if (idx != -1)
            this._onCloseHandlers.splice(idx, 1);
    }
    close() {
        this.client.write({ op: "close", session: this.sessionId });
        delete this.client.sessions[this.sessionId];
        this._onCloseHandlers.forEach(x => x(this));
    }
    clone() {
        return __awaiter(this, void 0, void 0, function* () {
            return new Promise((resolve, reject) => {
                let id = this.client.nextId;
                this.messageHandlers[id] = (msg) => {
                    let sess = new NReplSession(msg["new-session"], this.client);
                    resolve(sess);
                    return true;
                };
                this.client.write({ op: "clone", session: this.sessionId, id });
            });
        });
    }
    _response(data) {
        if (this.messageHandlers[data.id]) {
            let res = this.messageHandlers[data.id](data);
            if (res)
                delete this.messageHandlers[data.id];
        }
    }
    describe(verbose) {
        return new Promise((resolve, reject) => {
            let id = this.client.nextId;
            this.messageHandlers[id] = (msg) => {
                resolve(msg);
                return true;
            };
            this.client.write({ op: "describe", id: id, session: this.sessionId, verbose });
        });
    }
    listSessions() {
        return new Promise((resolve, reject) => {
            let id = this.client.nextId;
            this.messageHandlers[id] = (msg) => {
                resolve(msg);
                return true;
            };
            this.client.write({ op: "ls-sessions", id: id, session: this.sessionId });
        });
    }
    stacktrace() {
        return new Promise((resolve, reject) => {
            let id = this.client.nextId;
            this.messageHandlers[id] = (msg) => {
                resolve(msg);
                return true;
            };
            this.client.write({ op: "stacktrace", id, session: this.sessionId });
        });
    }
    eval(code, opts = {}) {
        let id = this.client.nextId;
        let evaluation = new NReplEvaluation(id, this, opts.stderr, opts.stdout, new Promise((resolve, reject) => {
            let ex;
            let value;
            this.messageHandlers[id] = (msg) => {
                if (msg.out)
                    evaluation.out(msg.out);
                if (msg.err)
                    evaluation.out(msg.err);
                if (msg.ns)
                    evaluation.ns = msg.ns;
                if (msg.ex)
                    ex = msg.ex;
                if (msg.value)
                    value = msg.value;
                if (msg["pprint-out"])
                    evaluation.pprintOut = msg["pprint-out"];
                if (msg.status && msg.status.indexOf("done") != -1) {
                    if (ex)
                        reject(ex);
                    else if (value)
                        resolve(value);
                    else if (evaluation.pprintOut)
                        resolve(evaluation.pprintOut);
                    else
                        resolve("");
                    return true;
                }
            };
            this.client.write(Object.assign({ op: "eval", session: this.sessionId, code, id }, opts));
        }));
        return evaluation;
    }
    interrupt(interruptId) {
        let id = this.client.nextId;
        return new Promise((resolve, reject) => {
            this.messageHandlers[id] = (msg) => {
                resolve();
                return true;
            };
            this.client.write({ op: "interrupt", session: this.sessionId, "interrupt-id": interruptId, id });
        });
    }
    stdin(message) {
        this.client.write({ op: "stdin", stdin: message, session: this.sessionId });
    }
    loadFile(file, opts = {}) {
        let id = this.client.nextId;
        let evaluation = new NReplEvaluation(id, this, opts.stderr, opts.stdout, new Promise((resolve, reject) => {
            this.messageHandlers[id] = (msg) => {
                if (msg.value)
                    resolve(msg.value);
                if (msg.out)
                    evaluation.out(msg.out);
                if (msg.err)
                    evaluation.out(msg.err);
                if (msg.ex) {
                    this.stacktrace().then(ex => reject(ex));
                }
                if (msg.status && msg.status.indexOf("done") != -1)
                    return true;
            };
            this.client.write({ op: "load-file", session: this.sessionId, file, id, "file-name": opts.fileName, "file-path": opts.filePath });
        }));
        return evaluation;
    }
    complete(ns, symbol, context) {
        return new Promise((resolve, reject) => {
            let id = this.client.nextId;
            this.messageHandlers[id] = (msg) => {
                resolve(msg);
                return true;
            };
            this.client.write({ op: "complete", ns, symbol, id, session: this.sessionId, context });
        });
    }
    info(ns, symbol) {
        return new Promise((resolve, reject) => {
            let id = this.client.nextId;
            this.messageHandlers[id] = (msg) => {
                resolve(msg);
                return true;
            };
            this.client.write({ op: "info", ns, symbol, id, session: this.sessionId });
        });
    }
    test(ns) {
        return new Promise((resolve, reject) => {
            let id = this.client.nextId;
            this.messageHandlers[id] = (msg) => {
                resolve(msg);
                return true;
            };
            this.client.write({
                op: "test-var-query", ns, id, session: this.sessionId, "var-query": {
                    "ns-query": {
                        exactly: [ns]
                    }
                }
            });
        });
    }
    testAll() {
        return new Promise((resolve, reject) => {
            let id = this.client.nextId;
            this.messageHandlers[id] = (msg) => {
                resolve(msg);
                return true;
            };
            this.client.write({ op: "test-all", id, session: this.sessionId, "load?": true });
        });
    }
    retest() {
        return new Promise((resolve, reject) => {
            let id = this.client.nextId;
            this.messageHandlers[id] = (msg) => {
                resolve(msg);
                return true;
            };
            this.client.write({ op: "retest", id, session: this.sessionId });
        });
    }
    loadAll() {
        return new Promise((resolve, reject) => {
            let id = this.client.nextId;
            this.messageHandlers[id] = (msg) => {
                resolve(msg);
                return true;
            };
            this.client.write({ op: "ns-load-all", id, session: this.sessionId });
        });
    }
    listNamespaces(regexps) {
        return new Promise((resolve, reject) => {
            let id = this.client.nextId;
            this.messageHandlers[id] = (msg) => {
                resolve(msg);
                return true;
            };
            this.client.write({ op: "ns-list", id, session: this.sessionId, "filter-regexps": regexps });
        });
    }
    nsPath(ns) {
        return new Promise((resolve, reject) => {
            let id = this.client.nextId;
            this.messageHandlers[id] = (msg) => {
                resolve(msg);
                return true;
            };
            this.client.write({ op: "ns-path", id, ns, session: this.sessionId });
        });
    }
    _refresh(cmd, opts = {}) {
        return new Promise((resolve, reject) => {
            let id = this.client.nextId;
            let reloaded = [];
            let error, errorNs, status, err = "";
            this.messageHandlers[id] = (msg) => {
                if (msg.reloading)
                    reloaded = msg.reloading;
                if (msg.status && msg.status.indexOf("ok") != -1)
                    status = "ok";
                if (msg.status && msg.status.indexOf("error") != -1) {
                    status = "error";
                    error = msg.error;
                    errorNs = msg["error-ns"];
                }
                if (msg.err)
                    err += msg.err;
                if (msg.status && msg.status.indexOf("done") != -1) {
                    let res = { reloaded, status };
                    if (error)
                        res.error = error;
                    if (errorNs)
                        res.errorNs = error;
                    if (err)
                        res.err = err;
                    resolve(res);
                    return true;
                }
            };
            this.client.write(Object.assign({ op: cmd, id, session: this.sessionId }, opts));
        });
    }
    refresh(opts = {}) {
        return this._refresh("refresh", opts);
    }
    refreshAll(opts = {}) {
        return this._refresh("refresh-all", opts);
    }
    formatCode(code, options) {
        return new Promise((resolve, reject) => {
            let id = this.client.nextId;
            this.messageHandlers[id] = (msg) => {
                resolve(msg);
                return true;
            };
            this.client.write({ op: "format-code", code, options });
        });
    }
}
exports.NReplSession = NReplSession;
/**
 * A running nREPL eval call.
 */
class NReplEvaluation {
    constructor(id, session, stderr, stdout, value) {
        this.id = id;
        this.session = session;
        this.stderr = stderr;
        this.stdout = stdout;
        this.value = value;
    }
    out(message) {
        if (this.stdout)
            this.stdout(message);
    }
    err(message) {
        if (this.stderr)
            this.stderr(message);
    }
    in(message) {
        this.session.stdin(message);
    }
    interrupt() {
        return this.session.interrupt(this.id);
    }
}
exports.NReplEvaluation = NReplEvaluation;
//# sourceMappingURL=index.js.map