"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const vscode = require("vscode");
const connector_1 = require("./connector");
const path = require("path");
const state = require("./state");
const status_1 = require("./status");
const fs_1 = require("fs");
// REPL
function activeReplWindow() {
    for (let w in replWindows) {
        if (replWindows[w].panel.active)
            return replWindows[w];
    }
}
exports.activeReplWindow = activeReplWindow;
class REPLWindow {
    constructor(panel, session, type) {
        this.panel = panel;
        this.session = session;
        this.type = type;
        this.useBuffer = false;
        this.buffer = [];
        this.onClose = () => this.postMessage({ type: "disconnected" });
        this.ns = "user";
        vscode.commands.executeCommand("setContext", "calva:pareditValid", true);
        this.initialized = new Promise((resolve, reject) => {
            this.panel.webview.onDidReceiveMessage((msg) => __awaiter(this, void 0, void 0, function* () {
                if (msg.type == "init") {
                    this.postMessage({ type: "init", ns: this.ns, history: state.extensionContext.workspaceState.get(this.type + "-history") || [] });
                    resolve();
                }
                if (msg.type == "history") {
                    let history = (state.extensionContext.workspaceState.get(this.type + "-history") || []);
                    history.push(msg.line);
                    state.extensionContext.workspaceState.update(this.type + "-history", history);
                }
                if (msg.type == "complete") {
                    let result = yield this.session.complete(this.ns, msg.symbol, msg.context);
                    this.postMessage({ type: "complete", data: result });
                }
                if (msg.type == "interrupt" && this.evaluation)
                    this.evaluation.interrupt();
                if (msg.type == "read-line") {
                    this.replEval(msg.line);
                }
                if (msg.type == "goto-file") {
                    vscode.workspace.openTextDocument(vscode.Uri.parse(msg.file)).then(d => {
                        let pos = new vscode.Position(msg.line - 1, 0);
                        vscode.window.showTextDocument(d, { viewColumn: vscode.ViewColumn.One, selection: new vscode.Range(pos, pos) });
                    });
                }
                if (msg.type == "info") {
                    let result = yield this.session.info(msg.ns, msg.symbol);
                    this.postMessage({ type: "info", data: result });
                }
            }));
        });
        this.panel.onDidDispose(e => {
            if (this.evaluation)
                this.evaluation.interrupt();
            delete replWindows[this.type];
            this.session.close();
            session.removeOnCloseHandler(this.onClose);
        });
        panel.onDidChangeViewState(e => {
            this.useBuffer = !e.webviewPanel.visible;
            vscode.commands.executeCommand("setContext", "calva:pareditValid", e.webviewPanel.active);
            if (e.webviewPanel.visible) {
                this.buffer.forEach(x => this.panel.webview.postMessage(x));
                this.buffer = [];
            }
            status_1.default.update();
        });
        panel.iconPath = vscode.Uri.file(path.join(ctx.extensionPath, "html", "/calva-icon.png"));
        let html = fs_1.readFileSync(path.join(ctx.extensionPath, "html/index.html")).toString();
        html = html.replace("{{baseUri}}", getUrl());
        html = html.replace("{{script}}", getUrl("/main.js"));
        html = html.replace("{{font}}", getUrl("/fira_code.css"));
        html = html.replace("{{logo}}", getUrl("/calva.png"));
        panel.webview.html = html;
        this.connect(session);
    }
    postMessage(msg) {
        if (this.useBuffer)
            this.buffer.push(msg);
        else
            this.panel.webview.postMessage(msg);
    }
    /**
     * Connects this repl window to the given session.
     *
     * @param session the session to connect to this repl window
     */
    connect(session) {
        return __awaiter(this, void 0, void 0, function* () {
            this.session = session;
            let res = this.session.eval("nil");
            yield res.value;
            this.ns = res.ns;
            session.addOnCloseHandler(this.onClose);
        });
    }
    evaluate(ns, text) {
        this.postMessage({ type: "do-eval", value: text, ns });
    }
    setNamespace(ns) {
        return __awaiter(this, void 0, void 0, function* () {
            this.postMessage({ type: "set-ns!", ns });
            this.ns = ns;
        });
    }
    replEval(line, ns) {
        return __awaiter(this, void 0, void 0, function* () {
            this.evaluation = this.session.eval(line, {
                stderr: m => this.postMessage({ type: "stderr", value: m }),
                stdout: m => this.postMessage({ type: "stdout", value: m })
            });
            try {
                this.postMessage({ type: "repl-response", value: yield this.evaluation.value, ns: this.ns = ns || this.evaluation.ns || this.ns });
            }
            catch (e) {
                this.postMessage({ type: "repl-error", ex: e });
                let stacktrace = yield this.session.stacktrace();
                this.postMessage({ type: "repl-ex", ex: JSON.stringify(stacktrace) });
            }
            this.evaluation = null;
        });
    }
    executeCommand(command) {
        this.panel.webview.postMessage({ type: "ui-command", value: command });
    }
}
let ctx;
let replWindows = {};
function getUrl(name) {
    if (name)
        return vscode.Uri.file(path.join(ctx.extensionPath, "html", name)).with({ scheme: 'vscode-resource' }).toString();
    else
        return vscode.Uri.file(path.join(ctx.extensionPath, "html")).with({ scheme: 'vscode-resource' }).toString();
}
function reconnectRepl(mode, session) {
    return __awaiter(this, void 0, void 0, function* () {
        if (replWindows[mode]) {
            yield replWindows[mode].connect(session);
            replWindows[mode].postMessage({ type: "reconnected", ns: replWindows[mode].ns });
        }
    });
}
exports.reconnectRepl = reconnectRepl;
function openReplWindow(mode = "clj") {
    return __awaiter(this, void 0, void 0, function* () {
        if (replWindows[mode]) {
            replWindows[mode].panel.reveal(vscode.ViewColumn.Two, true);
            return replWindows[mode];
        }
        let session = mode == "clj" ? connector_1.cljSession : connector_1.cljsSession;
        if (!session) {
            vscode.window.showErrorMessage("Not connected to nREPL");
            return;
        }
        session = yield session.clone();
        let title = mode == "clj" ? "CLJ REPL" : "CLJS Repl";
        const panel = vscode.window.createWebviewPanel("replInteractor", title, vscode.ViewColumn.Two, { retainContextWhenHidden: true, enableScripts: true, localResourceRoots: [vscode.Uri.file(path.join(ctx.extensionPath, 'html'))] });
        let repl = replWindows[mode] = new REPLWindow(panel, session, mode);
        yield repl.initialized;
        return repl;
    });
}
exports.openReplWindow = openReplWindow;
function activate(context) {
    ctx = context;
    context.subscriptions.push(vscode.commands.registerCommand('calva.openCljReplWindow', () => openReplWindow("clj")));
    context.subscriptions.push(vscode.commands.registerCommand('calva.openCljsReplWindow', () => openReplWindow("cljs")));
}
exports.activate = activate;
//# sourceMappingURL=repl-window.js.map